# Braintree-API for iOS

Braintree-API provides a native Objective-C interface that allows iOS app to communicate directly with Braintree.

It enables you to save and retrieve payment methods on Braintree via the Client API.

## Prerequisites

To get started, you will need:

* A Braintree account.
* A secure, authenticated server with a Braintree integration.


## How it works

Your app configures the Braintree iOS library via a Braintree **client token** generated on your server with a [Braintree server-side library](https://github.com/braintree/client-sdk-docs/blob/master/SERVER_DOCS.md).

1. Your app requests a Braintree **client token** from your server via a secure, authenticated endpoint.
2. Your server uses a [Braintree server-side library](https://github.com/braintree/client-sdk-docs/blob/master/SERVER_DOCS.md#creating-a-client-token) to generate a client-token and return it to your app.
3. Your app initializes the Braintree iOS library with the client token.

Once the Braintree iOS library is initialized, your app may use it to perform operations against Braintree resources to create and/or retrieve payment method nonces.

Your app will then submit a payment method nonce to your server, which will turn around and [do something useful with it](https://github.com/braintree/client-sdk-docs/blob/develop/SERVER_DOCS.md#using-a-payment-method-nonce).


## Setup

Use [CocoaPods](http://cocoapods.org/) to add `Braintree-API` to your project. In your `Podfile` add:

```
pod "Braintree/API"
```

If you've never used CocoaPods before, [this website](http://guides.cocoapods.org/using/getting-started.html) does a great job explaining what it is and how it works. We believe that dependency management will make iOS developers more productive and are curious about your feedback. (It is possible to pull in our code manually, but that is not recommended at this time.)


## Usage

### Client Token

The client token is generated by a server-side library and must be signed with your Braintree private key. The client token contains crucial configuration data, and this library cannot function without it. 

In test, we provide a mechanism to generate a fake client token locally. This client token, however, is only useful for evaluating this library. It causes the library to work *offline only*. As a result, no data is transmitted to Braintree, and transactions cannot be created with it.

### Obtain a Client Token

To get started, you'll need to generate a client token on your server and make it accessible to your client. 
(If you wish to try out this library without server integration, you can [skip this section](#offline-test-client-token))

An example server endpoint that returns the client token:

```
# Your server-side sinatra app
get '/client_token.json' do
  acts_as_authenticated
  customer_result = Braintree::Customer.create
  client_token = Braintree::ClientToken.generate(:customer_id => customer_result.customer.id)

  { :client_token => client_token }.to_json
end
```

It's up to you to communicate this to the client in real time as needed. For example, you can perform an authenticated HTTP request against your app's private API.

```
/* Your iOS App */
#import <AFNetworking/AFNetworking.h>
 __block NSString *CLIENT_TOKEN_FROM_SERVER;

AFHTTPRequestOperationManager *manager = [AFHTTPRequestOperationManager manager];
[manager GET:@"http://your.domain.com/client_token.json" parameters:nil success:^(AFHTTPRequestOperation *operation, id responseObject) {
    CLIENT_TOKEN_FROM_SERVER = responseObject[@"client_token"];
} failure:nil];
```

> What *is* that thing? You may note that the client token looks like JSON, but you can treat it as an opaque string that simply makes a roundtrip from your server-side Braintree Client library to the Braintree iOS Library.


### Initialize

Now that you have a client token, you're ready to start:

```
#import <Braintree/Braintree-API.h>
BTClient *braintree = [[BTClient alloc] initWithClientToken:CLIENT_TOKEN_FROM_SERVER];
```

#### "Offline Test" Client Token

You may want to try out our library without a server integration. To this end, we provide an offline mode, which emulates the actual API's behavior. Most of the library's semantics are identical in offline mode.

Each time you invoke `[BTClient offlineTestClientToken]`, the offline client's persistent data are reset.

In order to create an offline client, use this code:

```
#import <BTClient+Offline.h>
BTClient *testClient = [[BTClient alloc] initWithClientToken:[BTClient offlineTestClientToken]];
```

Behind the scenes, this client is a totally normal `BTClient`. The client token you obtain is initialized a special local-only Client API URL. We use [`NSURLProtocol`](http://nshipster.com/nsurlprotocol/) to intercept network calls with a particular custom URL scheme.

> Note: **Do not** ship apps with offline mock clients.


### Usage

Here's an example of saving a card:

```
[braintree saveCardWithNumber:@"4111111111111111"
              expirationMonth:@"12"
               expirationYear:@"2018"
                      success:^(BTCardPaymentMethod *card) {
                          NSLog(@"%@", card); // <BTCardPaymentMethod: number:4111 expirationMonth:12 expirationYear:2014 type:VISA nonce:1234>
                       }
                      failure:^(NSError *error){
                          NSLog(@"%@", error.domain); // BTCustomerInputErrorDomain
                       }];

```

See the [header files](./Braintree-API/Public/) for all supported `BTClient` operations.


### Summary

Phew! At this point you have a `BTClient` that is ready to go. You can now interact with Braintree resources like cards and PayPal accounts directly from the client.


## See Also

* [`docs/DESIGN.md`](./docs/DESIGN.md) offers a deeper look into the thought processes that went into this library.
* [`docs/TLDR.md`](./docs/TLDR.md) less than 500 bytes of guidance.
